@@ -2,7 +2,7 @@
  *
  * GDM - The GNOME Display Manager
  * Copyright (C) 1998, 1999, 2000 Martin K. Petersen <mkp@mkp.net>
- * Copyright (C) 2005 Brian Cameron <brian.cameron@sun.com>
+ * Copyright (C) 2005 Sun Microsystems, Inc.
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -71,6 +71,7 @@ static GSList *displays = NULL;
 static GSList *xservers = NULL;
 
 static gint high_display_num = 0;
+static const char *default_config_file = NULL;
 static char *custom_config_file = NULL;
 
 static uid_t GdmUserId;   /* Userid  under which gdm should run */
@@ -846,10 +847,9 @@ lookup_notify_key (GdmConfig  *config,
 	else if (is_key (keystring, GDM_KEY_TIMED_LOGIN))
 		nkey = g_strdup (GDM_NOTIFY_TIMED_LOGIN);
 	else if (strcmp (group, GDM_CONFIG_GROUP_CUSTOM_CMD) == 0 &&
-		 g_str_has_prefix (key, "CustomCommand") &&
-		 strlen (key) == 14) {
+		 g_str_has_prefix (key, "CustomCommand")) {
 		/* this should match 'CustomCommandN' */
-		nkey = g_strdup (key);
+		nkey = g_strdup (GDM_NOTIFY_CUSTOM_CMD_TEMPLATE);
 	}
 	g_free (keystring);
 
@@ -1228,101 +1228,6 @@ gdm_daemon_config_load_xservers (GdmConfig *config)
 	gdm_daemon_config_ensure_one_xserver (config);
 }
 
-/**
- * gdm_daemon_config_update_key
- *
- * Will cause a the GDM daemon to re-read the key from the configuration
- * file and cause notify signal to be sent to the slaves for the
- * specified key, if appropriate.
- * Obviously notification is not needed for configuration options only
- * used by the daemon.  This function is called when the UPDDATE_CONFIG
- * sockets command is called.
- *
- * To add a new notification, a GDM_NOTIFY_* argument will need to be
- * defined in gdm-daemon-config-keys.h, supporting logic placed in the
- * notify_cb function and in the gdm_slave_handle_notify function
- * in slave.c.
- */
-gboolean
-gdm_daemon_config_update_key (const char *keystring)
-{
-	gboolean              rc;
-	gboolean              res;
-	char                 *group;
-	char                 *key;
-	char                 *locale;
-	const GdmConfigEntry *entry;
-
-	rc = FALSE;
-	group = key = locale = NULL;
-
-	/*
-	 * Do not allow these keys to be updated, since GDM would need
-	 * additional work, or at least heavy testing, to make these keys
-	 * flexible enough to be changed at runtime.
-	 */
-	if (is_key (keystring, GDM_KEY_PID_FILE) ||
-	    is_key (keystring, GDM_KEY_CONSOLE_NOTIFY) ||
-	    is_key (keystring, GDM_KEY_USER) ||
-	    is_key (keystring, GDM_KEY_GROUP) ||
-	    is_key (keystring, GDM_KEY_LOG_DIR) ||
-	    is_key (keystring, GDM_KEY_SERV_AUTHDIR) ||
-	    is_key (keystring, GDM_KEY_USER_AUTHDIR) ||
-	    is_key (keystring, GDM_KEY_USER_AUTHFILE) ||
-	    is_key (keystring, GDM_KEY_USER_AUTHDIR_FALLBACK)) {
-		return FALSE;
-	}
-
-	/* reload backend files if necessary */
-	gdm_config_load (daemon_config, NULL);
-
-	/* Shortcut for updating all XDMCP parameters */
-	if (is_key (keystring, "xdmcp/PARAMETERS")) {
-		rc = gdm_daemon_config_update_key (GDM_KEY_DISPLAYS_PER_HOST);
-                if (rc == TRUE)
-			rc = gdm_daemon_config_update_key (GDM_KEY_MAX_PENDING);
-                if (rc == TRUE)
-			rc = gdm_daemon_config_update_key (GDM_KEY_MAX_WAIT);
-                if (rc == TRUE)
-			rc = gdm_daemon_config_update_key (GDM_KEY_MAX_SESSIONS);
-                if (rc == TRUE)
-			rc = gdm_daemon_config_update_key (GDM_KEY_INDIRECT);
-                if (rc == TRUE)
-			rc = gdm_daemon_config_update_key (GDM_KEY_MAX_INDIRECT);
-                if (rc == TRUE)
-			rc = gdm_daemon_config_update_key (GDM_KEY_MAX_WAIT_INDIRECT);
-                if (rc == TRUE)
-			rc = gdm_daemon_config_update_key (GDM_KEY_PING_INTERVAL);
-		goto out;
-	}
-
-	/* find the entry for the key */
-	res = gdm_common_config_parse_key_string (keystring,
-						  &group,
-						  &key,
-						  &locale,
-						  NULL);
-	if (! res) {
-		gdm_error ("Could not parse configuration key %s", keystring);
-		goto out;
-	}
-
-	entry = gdm_config_lookup_entry (daemon_config, group, key);
-	if (entry == NULL) {
-		gdm_error ("Request for invalid configuration key %s", keystring);
-		goto out;
-	}
-
-	rc = gdm_config_process_entry (daemon_config, entry, NULL);
-
- out:
-	g_free (group);
-	g_free (key);
-	g_free (locale);
-
-	return rc;
-}
-
 /**
  * check_logdir
  * check_servauthdir
@@ -1528,7 +1433,7 @@ gdm_daemon_config_load_displays (GdmConfig *config)
 		g_strfreev (value_list);
 	}
 
-	g_ptr_array_free ((GPtrArray*) keys, TRUE);
+	g_free (keys);
 }
 
 static gboolean
@@ -2213,6 +2118,139 @@ gdm_daemon_check_permissions (GdmConfig *config,
 	g_free (auth_path);
 }
 
+void
+gdm_daemon_load_config_file (GdmConfig **load_config)
+{
+	GError       *error;
+
+	if (*load_config == NULL)
+		*load_config = gdm_config_new ();
+
+	gdm_config_set_validate_func (*load_config, validate_cb, NULL);
+	gdm_config_add_static_entries (*load_config, gdm_daemon_config_entries);
+	gdm_config_set_default_file (*load_config, default_config_file);
+	gdm_config_set_custom_file (*load_config, custom_config_file);
+
+	/* load the data files */
+	error = NULL;
+	gdm_config_load (*load_config, &error);
+	if (error != NULL) {
+		gdm_error ("Unable to load configuration: %s", error->message);
+		g_error_free (error);
+	}
+
+	/* populate the database with all specified entries */
+	gdm_config_process_all (*load_config, &error);
+}
+
+/**
+ * gdm_daemon_config_update_key
+ *
+ * Will cause a the GDM daemon to re-read the key from the configuration
+ * file and cause notify signal to be sent to the slaves for the
+ * specified key, if appropriate.
+ * Obviously notification is not needed for configuration options only
+ * used by the daemon.  This function is called when the UPDDATE_CONFIG
+ * sockets command is called.
+ *
+ * To add a new notification, a GDM_NOTIFY_* argument will need to be
+ * defined in gdm-daemon-config-keys.h, supporting logic placed in the
+ * notify_cb function and in the gdm_slave_handle_notify function
+ * in slave.c.
+ */
+gboolean
+gdm_daemon_config_update_key (const char *keystring)
+{
+	const GdmConfigEntry *entry;
+	GdmConfigValue       *value;
+        GdmConfig            *temp_config;
+	gboolean              rc;
+	gboolean              res;
+	char                 *group;
+	char                 *key;
+	char                 *locale;
+
+	rc = FALSE;
+	group = key = locale = NULL;
+	temp_config = NULL;
+
+	/*
+	 * Do not allow these keys to be updated, since GDM would need
+	 * additional work, or at least heavy testing, to make these keys
+	 * flexible enough to be changed at runtime.
+	 */
+	if (is_key (keystring, GDM_KEY_PID_FILE) ||
+	    is_key (keystring, GDM_KEY_CONSOLE_NOTIFY) ||
+	    is_key (keystring, GDM_KEY_USER) ||
+	    is_key (keystring, GDM_KEY_GROUP) ||
+	    is_key (keystring, GDM_KEY_LOG_DIR) ||
+	    is_key (keystring, GDM_KEY_SERV_AUTHDIR) ||
+	    is_key (keystring, GDM_KEY_USER_AUTHDIR) ||
+	    is_key (keystring, GDM_KEY_USER_AUTHFILE) ||
+	    is_key (keystring, GDM_KEY_USER_AUTHDIR_FALLBACK)) {
+		return FALSE;
+	}
+
+	/* Load configuration file */
+	gdm_daemon_load_config_file (&temp_config);
+
+	if (is_key (keystring, "xservers/PARAMETERS")) {
+		gdm_daemon_config_load_xservers (temp_config);
+		goto out;
+	}
+
+	/* Shortcut for updating all XDMCP parameters */
+	if (is_key (keystring, "xdmcp/PARAMETERS")) {
+		rc = gdm_daemon_config_update_key (GDM_KEY_DISPLAYS_PER_HOST);
+                if (rc == TRUE)
+			rc = gdm_daemon_config_update_key (GDM_KEY_MAX_PENDING);
+                if (rc == TRUE)
+			rc = gdm_daemon_config_update_key (GDM_KEY_MAX_WAIT);
+                if (rc == TRUE)
+			rc = gdm_daemon_config_update_key (GDM_KEY_MAX_SESSIONS);
+                if (rc == TRUE)
+			rc = gdm_daemon_config_update_key (GDM_KEY_INDIRECT);
+                if (rc == TRUE)
+			rc = gdm_daemon_config_update_key (GDM_KEY_MAX_INDIRECT);
+                if (rc == TRUE)
+			rc = gdm_daemon_config_update_key (GDM_KEY_MAX_WAIT_INDIRECT);
+                if (rc == TRUE)
+			rc = gdm_daemon_config_update_key (GDM_KEY_PING_INTERVAL);
+		goto out;
+	}
+
+	/* find the entry for the key */
+	res = gdm_common_config_parse_key_string (keystring,
+						  &group,
+						  &key,
+						  &locale,
+						  NULL);
+	if (! res) {
+		gdm_error ("Could not parse configuration key %s", keystring);
+		goto out;
+	}
+
+	entry = gdm_config_lookup_entry (temp_config, group, key);
+	if (entry == NULL) {
+		gdm_error ("Request for invalid configuration key %s", keystring);
+		goto out;
+	}
+
+	rc = gdm_config_process_entry (temp_config, entry, NULL);
+
+	gdm_config_get_value_for_id (temp_config, entry->id, &value);
+	gdm_config_set_value_for_id (daemon_config, entry->id, value);
+
+ out:
+	if (temp_config != NULL)
+		gdm_config_free (temp_config);
+	g_free (group);
+	g_free (key);
+	g_free (locale);
+
+	return rc;
+}
+
 /**
  * gdm_daemon_config_parse
  *
@@ -2224,41 +2262,22 @@ gdm_daemon_config_parse (const char *config_file,
 {
 	uid_t         uid;
 	gid_t         gid;
-	GError       *error;
 	gboolean      xdmcp_enabled;
 	gboolean      dynamic_enabled;
 
-	displays          = NULL;
-	high_display_num  = 0;
+	displays            = NULL;
+	high_display_num    = 0;
 
 	/* Not NULL if config_file was set by command-line option. */
 	if (config_file == NULL) {
 		config_file = GDM_DEFAULTS_CONF;
 	}
-	custom_config_file = g_strdup (GDM_CUSTOM_CONF);
 
-	if (daemon_config == NULL)
-		daemon_config = gdm_config_new ();
+	default_config_file = config_file;
+	custom_config_file  = g_strdup (GDM_CUSTOM_CONF);
 
+	gdm_daemon_load_config_file (&daemon_config);
 	gdm_config_set_notify_func (daemon_config, notify_cb, NULL);
-	gdm_config_set_validate_func (daemon_config, validate_cb, NULL);
-
-	gdm_config_add_static_entries (daemon_config, gdm_daemon_config_entries);
-
-	gdm_config_set_default_file (daemon_config, config_file);
-	gdm_config_set_custom_file (daemon_config, custom_config_file);
-
-	/* load the data files */
-	error = NULL;
-	gdm_config_load (daemon_config, &error);
-	if (error != NULL) {
-		gdm_error ("Unable to load configuration: %s", error->message);
-		g_error_free (error);
-	}
-
-	/* populate the database with all specified entries */
-	gdm_config_process_all (daemon_config, &error);
-
 	gdm_daemon_config_load_xservers (daemon_config);
 
 	/* Only read the list if no_console is FALSE at this stage */
@@ -2868,7 +2887,7 @@ gdm_daemon_config_set_user_session_lang (gboolean savesess,
 	dmrc = gdm_common_config_load (cfgstr, NULL);
 	if (dmrc == NULL) {
 		gint fd = -1;
-		gdm_debug ("file: %s does not exist - creating it", cfgstr);
+		gdm_debug ("The user dmrc file %s does not exist - creating it", cfgstr);
 		VE_IGNORE_EINTR (fd = g_open (cfgstr,
 			O_CREAT | O_TRUNC | O_RDWR, 0644));
 
@@ -2880,7 +2899,7 @@ gdm_daemon_config_set_user_session_lang (gboolean savesess,
 		dmrc = gdm_common_config_load (cfgstr, NULL);
 
 		if (dmrc == NULL) {
-			gdm_debug ("failed to open %s after creating it", cfgstr);
+			gdm_debug ("Failed to open dmrc file %s after trying to create it", cfgstr);
 			return;
 		}
 	}
@@ -2891,8 +2910,10 @@ gdm_daemon_config_set_user_session_lang (gboolean savesess,
 
 	if (savelang) {
 		if (ve_string_empty (save_language)) {
-			/* we chose the system default language so wipe the
-			 * lang key */
+			/*
+			 * We chose the system default language so wipe the
+			 * lang key
+			 */
 			g_key_file_remove_key (dmrc, "Desktop", "Language", NULL);
 		} else {
 			g_key_file_set_string (dmrc, "Desktop", "Language", save_language);
@@ -2927,41 +2948,49 @@ gdm_daemon_config_get_user_session_lang (char      **usrsess,
 	cfg = gdm_common_config_load (cfgfile, NULL);
 	g_free (cfgfile);
 
-	save = FALSE;
+	save    = FALSE;
 	session = NULL;
-	lang = NULL;
+	lang    = NULL;
 
-	gdm_common_config_get_string (cfg, "Desktop/Session", &session, NULL);
-	if (session == NULL) {
+	if (cfg == NULL) {
 		session = g_strdup ("");
-	}
+		lang    = g_strdup ("");
+	} else {
+		gdm_common_config_get_string (cfg, "Desktop/Session", &session, NULL);
+		if (session == NULL) {
+			session = g_strdup ("");
+		}
 
-	/* this is just being truly anal about what users give us, and in case
-	 * it looks like they may have included a path whack it. */
-	p = strrchr (session, '/');
-	if (p != NULL) {
-		char *tmp = g_strdup (p + 1);
-		g_free (session);
-		session = tmp;
-	}
+		/*
+		 * This is just being truly anal about what users give us, and in case
+		 * it looks like they may have included a path whack it.
+		 */
+		p = strrchr (session, '/');
+		if (p != NULL) {
+			char *tmp = g_strdup (p + 1);
+			g_free (session);
+			session = tmp;
+		}
 
-	/* ugly workaround for migration */
-	if (strcmp (session, "Default") == 0 ||
-	    strcmp (session, "Default.desktop") == 0) {
-		g_free (session);
-		session = g_strdup ("default");
-		save = TRUE;
-	}
+		/* Ugly workaround for migration */
+		if (strcmp (session, "Default") == 0 ||
+		    strcmp (session, "Default.desktop") == 0) {
+			g_free (session);
+			session = g_strdup ("default");
+			save    = TRUE;
+		}
 
-	gdm_common_config_get_string (cfg, "Desktop/Language", &lang, NULL);
-	if (lang == NULL) {
-		lang = g_strdup ("");
+		gdm_common_config_get_string (cfg, "Desktop/Language", &lang, NULL);
+		if (lang == NULL) {
+			lang = g_strdup ("");
+		}
+
+		g_key_file_free (cfg);
 	}
 
 	if (usrsess != NULL) {
 		*usrsess = g_strdup (session);
 	}
-	g_free (session);
 
 	if (savesess != NULL) {
 		*savesess = save;
@@ -2970,7 +2999,7 @@ gdm_daemon_config_get_user_session_lang (char      **usrsess,
 	if (usrlang != NULL) {
 		*usrlang = g_strdup (lang);
 	}
-	g_free (lang);
 
-	g_key_file_free (cfg);
+	g_free (session);
+	g_free (lang);
 }
